"""Client library for using OAuth2, with applications without web access."""

__author__ = 'Luis Maia <luis.maia@xfel.eu>'
__version__ = '6.1.0'

from .oauth2_client_backend import Oauth2ClientBackend



"""Oauth2ClientBackend unitary tests"""

import unittest
from datetime import datetime, timedelta
from http import HTTPStatus

from freezegun import freeze_time

from oauth2_xfel_client.oauth2_client_backend import \
    Oauth2ClientBackend as Oauth2Client
from .common.secrets import *

__author__ = 'Luis Maia <luis.maia@xfel.eu>'


class Oauth2ClientBackendTest(unittest.TestCase):
    def setUp(self):
        self.client_id = CLIENT_OAUTH2_INFO['CLIENT_ID']
        self.client_secret = CLIENT_OAUTH2_INFO['CLIENT_SECRET']
        self.scope = None
        self.auth_url = CLIENT_OAUTH2_INFO['AUTH_URL']
        self.token_url = CLIENT_OAUTH2_INFO['TOKEN_URL']
        self.refresh_url = self.token_url

        self.base_api_url = BASE_API_URL

        self.valid_token = {}
        self.invalid_token = {
            u'access_token': USER_ACCESS_TOKEN,
            u'expires_at': 1410368490.381425,
            u'expires_in': 7200,
            u'token_type': u'bearer'
        }

    @staticmethod
    def validate_token_duration(duration_s, expected=7200):
        """Check that token duration is 2 hours or just under

        Token duration should be 2 hours, but it's sometimes slightly less,
        presumably because of a delay somewhere. Allow up to 5 seconds less.
        """
        assert duration_s <= expected, \
            "Token duration ({}) should be <= {}".format(duration_s, expected)

    def test_new_session(self):
        # No session token
        client = Oauth2Client(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            token_url=self.token_url,
            refresh_url=self.refresh_url,
            auth_url=self.auth_url,
        )
        tok = client.get_session_token()

        assert client.is_session_token_dt_valid(tok)
        assert not client.is_session_token_dt_valid(
            tok, datetime.now() + timedelta(seconds=7205)
        ), "Invalid session token expiring date (expect: 2h:5s in future)"

        self.assertEqual(tok['token_type'].lower(), 'bearer')
        self.assertEqual(client.headers['Authorization'],
                         'Bearer ' + tok['access_token'])

        # Token still valid, so this shouldn't create a new session
        client.check_session_token()
        self.assertEqual(client.get_session_token(), tok)

    def test_reuse_invalid_token(self):
        # Passing an invalid_token to the constructor (in param session_token)
        # will create a new token (the same happen when session_token=None)
        client = Oauth2Client(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            token_url=self.token_url,
            refresh_url=self.refresh_url,
            auth_url=self.auth_url,
            session_token=self.invalid_token
        )

        current_token = client.get_session_token()

        # Checking invalid_token
        assert not client.is_session_token_dt_valid(self.invalid_token)

        # Checking current_token != invalid_token
        self.assertNotEqual(self.invalid_token['access_token'],
                            current_token['access_token'],
                            "Invalid token was successfully used")

        self.assertEqual(client.oauth_token['access_token'],
                         current_token['access_token'],
                         "oauth_token and current_token must be equal")

        self.assertNotEqual(self.invalid_token['expires_at'],
                            current_token['expires_at'],
                            "Invalid token expires_at was successfully used")
        self.assertEqual(
            client.oauth_token['expires_at'],
            datetime.fromtimestamp(current_token['expires_at']),
            "oauth_token and current_token 'expires_at' must be equal")

        self.validate_token_duration(current_token['expires_in'])

        self.assertEqual(current_token['token_type'].lower(),
                         u'bearer',
                         "Token type is 'bearer'")

        assert client.is_session_token_valid(), "Invalid session token"

        assert client.is_session_token_dt_valid(
            current_token
        ), "Invalid session token expiring date (expect: now)"

        assert not client.is_session_token_dt_valid(
            current_token,
            datetime.now() + timedelta(seconds=7205)
        ), "Invalid session token expiring date (expect: 2h:5s in future)"

    def test_reuse_valid_token(self):
        # Get a valid token
        client1 = Oauth2Client(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            token_url=self.token_url,
            refresh_url=self.refresh_url,
            auth_url=self.auth_url,
        )
        c1_token = client1.get_session_token()

        # Passing a valid session_token should use it
        client2 = Oauth2Client(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            token_url=self.token_url,
            refresh_url=self.refresh_url,
            auth_url=self.auth_url,
            session_token=c1_token
        )
        c2_token = client2.get_session_token()

        self.assertEqual(c2_token, c1_token,
                         "Valid session token was not used")

        self.validate_token_duration(c1_token['expires_in'])
        self.validate_token_duration(c2_token['expires_in'])

        self.assertEqual(c1_token['token_type'].lower(), 'bearer')

        self.assertEqual(c2_token['token_type'].lower(), 'bearer')

        assert client1.is_session_token_valid(), "Invalid session token"

        assert client1.is_session_token_dt_valid(
            c2_token
        ), "Invalid session token expiring date (expect: now)"

        assert not client1.is_session_token_dt_valid(
            c2_token,
            dt=datetime.now() + timedelta(seconds=7205)
        ), "Invalid session token expiring date (expect: 2h:5s in future)"

    def test_request(self):
        client = Oauth2Client(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            token_url=self.token_url,
            refresh_url=self.refresh_url,
            auth_url=self.auth_url,
        )

        # .../api/me is a typical endpoint for the current EuXFEL services
        url = BASE_API_URL + 'me'
        headers = {
            'Accept': 'application/json; version=1',
            'X-User-Email': CLIENT_OAUTH2_INFO['EMAIL'],
        }
        r = client.session.get(url, headers=headers)
        self.assertEqual(r.status_code, HTTPStatus.OK)
        j = r.json()
        self.assertEqual(j['email'], CLIENT_OAUTH2_INFO['EMAIL'])

    def test_auto_renew(self):
        client = Oauth2Client(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            token_url=self.token_url,
            refresh_url=self.refresh_url,
            auth_url=self.auth_url,
        )
        orig_token = client.get_session_token()

        # .../api/me is a typical endpoint for the current EuXFEL services
        url = BASE_API_URL + 'me'
        headers = {
            'Accept': 'application/json; version=1',
            'X-User-Email': CLIENT_OAUTH2_INFO['EMAIL'],
        }

        # Tokens last 2 hours, so 3 hours in the future it will have expired.
        # This works because oauthlib checks the token expiry on the client
        with freeze_time(datetime.now() + timedelta(hours=3)):
            r = client.session.get(url, headers=headers)

        self.assertEqual(r.status_code, HTTPStatus.OK)

        # We should have got a new token
        self.assertNotEqual(client.get_session_token(), orig_token)


if __name__ == '__main__':
    unittest.main()



import logging
import time
from datetime import datetime

from oauthlib.oauth2 import BackendApplicationClient, TokenExpiredError
from requests import Timeout
from requests_oauthlib.oauth2_session import OAuth2Session

__author__ = 'Luis Maia <luis.maia@xfel.eu>'

log = logging.getLogger(__name__)

# Number of retries allowed before deal with the consequences!
DEF_MAX_RETRIES = 3

# Default Timeout for Request Connection and Read
DEF_TIMEOUT = 12


# requests_oauthlib supports automatically *refreshing* a token, an OAuth
# feature that doesn't require going through the full flow again.
# The XFEL services don't seem to allow this, but because we are using
# the backend flow, with no user interaction, it's relatively easy to get
# a brand new token.
# We are borrowing the parent class' auto_refresh_kwargs attribute to use
# for this purpose. refresh_token is disabled to avoid confusion.
class AutoRenewingOAuthSession(OAuth2Session):
    """OAuth session which automatically gets a new token on expiry
    """

    def request(self, method, url, *args, **kwargs):
        # If a previous token request failed, we may have no token.
        # Ensure we have one before making a request.
        if (not self.token) and url != self.auto_refresh_kwargs['token_url']:
            log.warning("Missing Oauth token - requesting one")
            self.fetch_token(**self.auto_refresh_kwargs)

        try:
            return super().request(method, url, *args, **kwargs)
        except TokenExpiredError:
            log.info("Expired Oauth token - requesting new one")
            self.fetch_token(**self.auto_refresh_kwargs)
            return super().request(method, url, *args, **kwargs)

    def refresh_token(self, *args, **kwargs):
        raise NotImplementedError(
            "Get a new token with fetch_token instead of refreshing it"
        )

    def fetch_token(self, max_retries=DEF_MAX_RETRIES, **kwargs):
        retries = max_retries
        retry_delay = 0
        while True:
            try:
                super().fetch_token(**kwargs)

            except Timeout as exp:
                # log.debug('Token request Timeout', exc_info=True)
                exp_msg = 'Exception [try {0}/{1} | timeout: {2} | ' \
                          'delay: {3}]: {4}'.format(retries,
                                                    max_retries,
                                                    kwargs['timeout'],
                                                    retry_delay, exp)
                log.debug(exp_msg, exc_info=True)
                # print('Timeout:' + '>' * 10 + exp_msg + '<' * 10)
                retries -= 1
                if retries <= 0:
                    raise

                # Sleep before retrying
                # (e.g. max_retries == 3)
                # Try 1: retry_delay = N.A.
                # Try 2: retry_delay = (3-1)/2 = 1 sec
                # Try 3: retry_delay = (3-1)/1 = 2 sec
                retry_delay = (max_retries - 1) / retries
                time.sleep(retry_delay)

                # # Increase timeout before retrying (FOR TESTING)
                # #
                # # (e.g. kwargs['max_retries'] == 3)
                # # Try 1: timeout (specified == ? || default == 3)
                # # Try 2: timeout = (3*4)/2 = 6 sec
                # # Try 3: timeout = (3*4)/1 = 12 sec
                # retry_timeout = (kwargs['max_retries'] * 4) / retries
                # kwargs['timeout'] = max(kwargs['timeout'], retry_timeout)

            except Exception as exp:
                exp_msg = 'Exception: {0}'.format(exp)
                log.debug(exp_msg, exc_info=True)
                # print('General:' + '-' * 10 + exp_msg + '-' * 10)
                raise

            else:
                log.debug('Got a new session token successfully')
                break


class Oauth2ClientBackend(object):
    # Parameter that will store the session used to "invoke" the APIs
    session = None

    def __init__(self, client_id, client_secret, scope, token_url,
                 refresh_url=None, auth_url=None, session_token=None,
                 max_retries=DEF_MAX_RETRIES,
                 timeout=DEF_TIMEOUT,
                 ssl_verify=True):

        self.client_secret = client_secret
        self.scope = scope
        self.token_url = token_url
        self.timeout = timeout  # Only applies to getting an Oauth token
        self.max_retries = max_retries

        # Not currently in use in this Oauth2 Strategy:
        # self.refresh_url = refresh_url
        # self.auth_url = auth_url

        # REQUESTS::ssl_verify (when endpoint uses HTTPS protocol)
        #
        # 1. Verify SSL Certificate only if `ssl_verify=True` (default)
        #    * Retro-compatible with ond code:
        #      - self.ssl_verify = token_url.startswith('https://in.xfel.eu')
        #    * Valid Certificate is expected (e.g. 'https://in.xfel.eu')
        #
        # 2. Self-signed endpoints ssl_verify must be set `ssl_verify=False`
        #    because only certificates issued by a certificate authority (CA)
        #    will be accepted by REQUESTS (e.g. 'https://127.0.0.1/')
        self.ssl_verify = ssl_verify

        self.fetch_token_kwargs = dict(
            token_url=self.token_url,
            client_secret=self.client_secret,
            max_retries=self.max_retries,
            timeout=self.timeout,
            verify=self.ssl_verify
        )

        # Configure client using "Backend Application Flow" Oauth 2.0 strategy
        self.client = BackendApplicationClient(client_id)

        # Negotiate with the server and obtains a valid session_token
        # after this self.session can be used to 'invoke' APIs
        self.auth_session(session_token=session_token)

    def auth_session(self, session_token=None):
        # If a session token was passed in & it's still valid,
        # create a session using it.
        if self.is_session_token_dt_valid(session_token):
            self._re_used_existing_session_token(session_token)
            # Nothing to do. Current Token is still valid!
            pass

        # Otherwise, try to get a new session token.
        else:
            try:
                log.debug('Will try to create a new session token')
                self._create_new_session_token()

            except Exception as exp:
                exp_msg = 'Exception: {0}'.format(exp)
                log.debug(exp_msg, exc_info=True)
                # print('New session:' + '+' * 10 + exp_msg + '+' * 10)
                raise

            else:
                log.debug('Got a new session token successfully')

        return True

    def _re_used_existing_session_token(self, session_token):
        self.session = AutoRenewingOAuthSession(
            client=self.client,
            scope=self.scope,
            token=session_token,
            auto_refresh_kwargs=self.fetch_token_kwargs,
        )

    def _create_new_session_token(self):
        self.session = AutoRenewingOAuthSession(
            client=self.client,
            scope=self.scope,
            auto_refresh_kwargs=self.fetch_token_kwargs,
        )

        self.session.fetch_token(**self.fetch_token_kwargs)

    def check_session_token(self):
        if not self.is_session_token_valid():
            self.auth_session()  # Get a new session & a new token

    def get_session_token(self):
        return self.session.token

    @property
    def headers(self):
        # This isn't needed - the session automatically adds this header to
        # requests. But downstream code may expect it to exist.
        auth_token_val = 'Bearer ' + self.session.token['access_token']
        return {'Authorization': auth_token_val}

    @property
    def oauth_token(self):
        tok = self.session.token.copy()
        # Convert expires_at to a datetime object, keeping previous interface
        tok['expires_at'] = datetime.fromtimestamp(tok['expires_at'])
        return tok

    def is_session_token_valid(self):
        current_token = self.get_session_token()
        return Oauth2ClientBackend.is_session_token_dt_valid(current_token)

    @staticmethod
    def is_session_token_dt_valid(session_token, dt=None):
        # Check session_token hash
        if session_token and 'expires_at' in session_token:
            # Convert Unix timestamp (seconds from the epoch) to datetime
            expires_dt = datetime.fromtimestamp(session_token['expires_at'])

            if dt is None:
                dt = datetime.now()

            # return True:
            # 1) If expire datetime is in future (token is still valid)
            # return False:
            # 1) If expire datetime is in past (a new token must be generated)
            return expires_dt > dt

        return False
